/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.domain;

import java.time.Duration;
import java.time.LocalDateTime;

/**
 * die Repräsentation eines Staus innerhalb der Stauprognose eine Baustelle.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class PrognostizierterStau {

    private final LocalDateTime startZeit;
    private LocalDateTime endZeit;

    private long maximaleLaenge;
    private LocalDateTime maximaleLaengeZeit;

    private Duration verlustZeit = Duration.ZERO;

    /**
     * erzeugt eine Instanz des Stauobjekts.
     *
     * @param zs          der Zeitpunkt der Entstehung des Staus
     * @param laenge      die Länge des Staus bei der Entstehung
     * @param verlustZeit die Verlustzeit zum Entstehungszeitpunkt
     */
    public PrognostizierterStau(final LocalDateTime zs, final long laenge, final Duration verlustZeit) {
        this.startZeit = zs;
        this.endZeit = zs;
        aktualisieren(zs, laenge, verlustZeit);
    }

    /**
     * setzt die Verlustzeit des Staus.<br>
     * Das Stauobjekt speichert die maximale Verlustzeit im Verlauf der Existenz des
     * Staus. Der übergebene Wert wird deshalb nur übernommen, wenn der bereits
     * gespeicherte kleiner ist.
     *
     * @param zeit die neu zu setzende Verlustzeit
     */
    private void setVerlustZeit(final Duration zeit) {
        if (isBeendet()) {
            throw new IllegalStateException(
                    "Es wurde versucht die Verlustzeit an einem beendeten Stauobjekt zu setzen");
        }
        if (zeit.compareTo(verlustZeit) > 0) {
            verlustZeit = zeit;
        }
    }

    /**
     * ermittelt, ob der Stau beendet wurde.<br>
     * Ein Stau gilt als beendet, wenn er eine Dauer &gt; 0 hat, d.h. wenn de
     * Endzeit größer als die Anfangszeit ist.
     *
     * @return true, wenn der Stau beendet ist
     */
    private boolean isBeendet() {
        return endZeit.isAfter(startZeit);
    }

    /**
     * setzt die Länge des Staus und den dazugehörigen Zeitpunkt.<br>
     * Das Stauobjekt speichert die maximale Länge im Verlauf der Existenz des
     * Staus. Der übergebene Wert wird deshalb nur übernommen, wenn der bereits
     * gespeicherte kleiner ist.
     *
     * @param zs     der Zeitpunkt zu dem die übergeben Steulänge erreicht wurde
     * @param laenge die Länge
     */
    private void setMaximaleLaenge(final LocalDateTime zs, final long laenge) {
        if (isBeendet()) {
            throw new IllegalStateException("Es wurde versucht die Staulänge an einem beendeten Stauobjekt zu setzen");
        }
        if (maximaleLaenge < Math.max(0, laenge)) {
            maximaleLaengeZeit = zs;
            maximaleLaenge = laenge;
        }
    }

    /**
     * liefert den Anfangszeitpunkt des Staus.
     *
     * @return den Zeitpunkt
     */
    public final LocalDateTime getStartZeit() {
        return startZeit;
    }

    /**
     * liefert die Dauer des Staus in Millisekunden.<br>
     * Wenn kein Endzeitpunkt bestimmt wurde, wird der Wert 0 geliefert.
     *
     * @return den Zeitpunkt oder 0
     */
    public final Duration getDauer() {
        return Duration.between(startZeit, endZeit);
    }

    /**
     * liefert die maximale Länge, die während der Existenz des Staus aufgetreten
     * ist.
     *
     * @return die Länge
     */
    public final long getMaxLaenge() {
        return maximaleLaenge;
    }

    /**
     * liefert den Zeitpunkt zu dem der Stau seine maximale Länge erreicht hat.
     *
     * @return den Zeitpunkt
     */
    public final LocalDateTime getMaxLaengeZeit() {
        return maximaleLaengeZeit;
    }

    public final Duration getVerlustZeit() {
        return verlustZeit;
    }

    /**
     * setzt die Endezeit und markiert damit den Stau als beendet.
     *
     * @param zs der Zeitpunkt für das Stauende
     */
    public final void beenden(final LocalDateTime zs) {
        endZeit = zs;
    }

    /**
     * aktualisiert die Daten des Stauobjekts.<br>
     * Die Wert für Länge und Verlustzeit werden entsprechend in das Stauobjekt
     * eingebracht.
     *
     * @param zs     der zugehörige Zeitpunkt
     * @param laenge die neue Staulänge
     * @param zeit   die neue Verlustzeit
     */
    public final void aktualisieren(final LocalDateTime zs, final long laenge, final Duration zeit) {
        setMaximaleLaenge(zs, laenge);
        setVerlustZeit(zeit);
    }

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("PrognostizierterStau [startZeit=").append(startZeit).append(", endZeit=").append(endZeit)
				.append(", maximaleLaenge=").append(maximaleLaenge).append(", maximaleLaengeZeit=")
				.append(maximaleLaengeZeit).append(", verlustZeit=").append(verlustZeit).append("]");
		return builder.toString();
	}
    
    
}
